@@ -36,13 +36,16 @@ import datetime
 import threading
 import warnings
 import weakref
+
 from collections import defaultdict
 
+from bson.codec_options import DEFAULT_CODEC_OPTIONS
 from bson.py3compat import (integer_types,
                             string_type)
 from bson.son import SON
 from pymongo import (common,
                      database,
+                     helpers,
                      message,
                      periodic_executor,
                      uri_parser)
@@ -63,6 +66,7 @@ from pymongo.server_type import SERVER_TYPE
 from pymongo.topology import Topology
 from pymongo.topology_description import TOPOLOGY_TYPE
 from pymongo.settings import TopologySettings
+from pymongo.write_concern import WriteConcern
 
 
 class MongoClient(common.BaseObject):
@@ -103,11 +107,12 @@ class MongoClient(common.BaseObject):
           read :ref:`multiprocessing` first.
 
         :Parameters:
-          - `host` (optional): hostname or IP address of the
-            instance to connect to, or a mongodb URI, or a list of
+          - `host` (optional): hostname or IP address of a single mongod or
+            mongos instance to connect to, or a mongodb URI, or a list of
             hostnames / mongodb URIs. If `host` is an IPv6 literal
             it must be enclosed in '[' and ']' characters following
-            the RFC2732 URL syntax (e.g. '[::1]' for localhost)
+            the RFC2732 URL syntax (e.g. '[::1]' for localhost). Multihomed
+            and round robin DNS addresses are **not** supported.
           - `port` (optional): port number on which to connect
           - `document_class` (optional): default class to use for
             documents returned from queries on this client
@@ -121,10 +126,16 @@ class MongoClient(common.BaseObject):
 
           | **Other optional parameters can be passed as keyword arguments:**
 
-          - `maxPoolSize` (optional): The maximum number of connections
-            that the pool will open simultaneously. If this is set, operations
-            will block if there are `maxPoolSize` outstanding connections
-            from the pool. Defaults to 100. Cannot be 0.
+          - `maxPoolSize` (optional): The maximum allowable number of
+            concurrent connections to each connected server. Requests to a
+            server will block if there are `maxPoolSize` outstanding
+            connections to the requested server. Defaults to 100. Cannot be 0.
+          - `minPoolSize` (optional): The minimum required number of concurrent
+            connections that the pool will maintain to each connected server.
+            Default is 0.
+          - `maxIdleTimeMS` (optional): The maximum number of milliseconds that
+            a connection can remain idle in the pool before being removed and
+            replaced. Defaults to `None` (no limit).
           - `socketTimeoutMS`: (integer or None) Controls how long (in
             milliseconds) the driver will wait for a response after sending an
             ordinary (non-monitoring) database operation before concluding that
@@ -148,6 +159,9 @@ class MongoClient(common.BaseObject):
           - `socketKeepAlive`: (boolean) Whether to send periodic keep-alive
             packets on connected sockets. Defaults to ``False`` (do not send
             keep-alive packets).
+          - `heartbeatFrequencyMS`: (optional) The number of milliseconds
+            between periodic server checks, or None to accept the default
+            frequency of 10 seconds.
           - `event_listeners`: a list or tuple of event listeners. See
             :mod:`~pymongo.monitoring` for details.
 
@@ -192,29 +206,37 @@ class MongoClient(common.BaseObject):
 
           - `ssl`: If ``True``, create the connection to the server using SSL.
             Defaults to ``False``.
+          - `ssl_certfile`: The certificate file used to identify the local
+            connection against mongod. Implies ``ssl=True``. Defaults to
+            ``None``.
           - `ssl_keyfile`: The private keyfile used to identify the local
             connection against mongod.  If included with the ``certfile`` then
             only the ``ssl_certfile`` is needed.  Implies ``ssl=True``.
             Defaults to ``None``.
-          - `ssl_certfile`: The certificate file used to identify the local
-            connection against mongod. Implies ``ssl=True``. Defaults to
-            ``None``.
+          - `ssl_pem_passphrase`: The password or passphrase for decrypting
+            the private key in ``ssl_certfile`` or ``ssl_keyfile``. Only
+            necessary if the private key is encrypted. Only supported by python
+            2.7.9+ (pypy 2.5.1+) and 3.3+. Defaults to ``None``.
           - `ssl_cert_reqs`: Specifies whether a certificate is required from
             the other side of the connection, and whether it will be validated
             if provided. It must be one of the three values ``ssl.CERT_NONE``
-            (certificates ignored), ``ssl.CERT_OPTIONAL``
-            (not required, but validated if provided), or ``ssl.CERT_REQUIRED``
-            (required and validated). If the value of this parameter is not
-            ``ssl.CERT_NONE`` and a value is not provided for ``ssl_ca_certs``
-            PyMongo will attempt to load system provided CA certificates.
-            If the python version in use does not support loading system CA
-            certificates then the ``ssl_ca_certs`` parameter must point
-            to a file of CA certificates. Implies ``ssl=True``. Defaults to
+            (certificates ignored), ``ssl.CERT_REQUIRED`` (certificates
+            required and validated), or ``ssl.CERT_OPTIONAL`` (the same as
+            CERT_REQUIRED, unless the server was configured to use anonymous
+            ciphers). If the value of this parameter is not ``ssl.CERT_NONE``
+            and a value is not provided for ``ssl_ca_certs`` PyMongo will
+            attempt to load system provided CA certificates. If the python
+            version in use does not support loading system CA certificates
+            then the ``ssl_ca_certs`` parameter must point to a file of CA
+            certificates. Implies ``ssl=True``. Defaults to
             ``ssl.CERT_REQUIRED`` if not provided and ``ssl=True``.
           - `ssl_ca_certs`: The ca_certs file contains a set of concatenated
             "certification authority" certificates, which are used to validate
             certificates passed from the other end of the connection.
             Implies ``ssl=True``. Defaults to ``None``.
+          - `ssl_crlfile`: The path to a PEM or DER formatted certificate
+            revocation list. Only supported by python 2.7.9+ (pypy 2.5.1+)
+            and 3.4+. Defaults to ``None``.
           - `ssl_match_hostname`: If ``True`` (the default), and
             `ssl_cert_reqs` is not ``ssl.CERT_NONE``, enables hostname
             verification using the :func:`~ssl.match_hostname` function from
@@ -222,6 +244,15 @@ class MongoClient(common.BaseObject):
             this to ``False`` as that could make your application vulnerable to
             man-in-the-middle attacks.
 
+          | **Read Concern options:**
+          | (If not set explicitly, this will use the server default)
+
+          - `readConcernLevel`: (string) The read concern level specifies the
+            level of isolation for read operations.  For example, a read
+            operation using a read concern level of ``majority`` will only
+            return data that has been written to a majority of nodes. If the
+            level is left unspecified, the server default will be used.
+
         .. mongodoc:: connections
 
         .. versionchanged:: 3.0
@@ -334,17 +365,19 @@ class MongoClient(common.BaseObject):
 
         self.__default_database_name = dbase
         self.__lock = threading.Lock()
-        self.__cursor_manager = CursorManager(self)
+        self.__cursor_manager = None
         self.__kill_cursors_queue = []
 
         self._event_listeners = options.pool_options.event_listeners
 
         # Cache of existing indexes used by ensure_index ops.
         self.__index_cache = {}
+        self.__index_cache_lock = threading.Lock()
 
         super(MongoClient, self).__init__(options.codec_options,
                                           options.read_preference,
-                                          options.write_concern)
+                                          options.write_concern,
+                                          options.read_concern)
 
         self.__all_credentials = {}
         creds = options.credentials
@@ -359,7 +392,8 @@ class MongoClient(common.BaseObject):
             monitor_class=monitor_class,
             condition_class=condition_class,
             local_threshold_ms=options.local_threshold_ms,
-            server_selection_timeout=options.server_selection_timeout)
+            server_selection_timeout=options.server_selection_timeout,
+            heartbeat_frequency=options.heartbeat_frequency)
 
         self._topology = Topology(self._topology_settings)
         if connect:
@@ -369,7 +403,7 @@ class MongoClient(common.BaseObject):
             client = self_ref()
             if client is None:
                 return False  # Stop the executor.
-            MongoClient._process_kill_cursors_queue(client)
+            MongoClient._process_periodic_tasks(client)
             return True
 
         executor = periodic_executor.PeriodicExecutor(
@@ -420,27 +454,29 @@ class MongoClient(common.BaseObject):
         """Test if `index` is cached."""
         cache = self.__index_cache
         now = datetime.datetime.utcnow()
-        return (dbname in cache and
-                coll in cache[dbname] and
-                index in cache[dbname][coll] and
-                now < cache[dbname][coll][index])
+        with self.__index_cache_lock:
+            return (dbname in cache and
+                    coll in cache[dbname] and
+                    index in cache[dbname][coll] and
+                    now < cache[dbname][coll][index])
 
     def _cache_index(self, dbname, collection, index, cache_for):
         """Add an index to the index cache for ensure_index operations."""
         now = datetime.datetime.utcnow()
         expire = datetime.timedelta(seconds=cache_for) + now
 
-        if database not in self.__index_cache:
-            self.__index_cache[dbname] = {}
-            self.__index_cache[dbname][collection] = {}
-            self.__index_cache[dbname][collection][index] = expire
+        with self.__index_cache_lock:
+            if database not in self.__index_cache:
+                self.__index_cache[dbname] = {}
+                self.__index_cache[dbname][collection] = {}
+                self.__index_cache[dbname][collection][index] = expire
 
-        elif collection not in self.__index_cache[dbname]:
-            self.__index_cache[dbname][collection] = {}
-            self.__index_cache[dbname][collection][index] = expire
+            elif collection not in self.__index_cache[dbname]:
+                self.__index_cache[dbname][collection] = {}
+                self.__index_cache[dbname][collection][index] = expire
 
-        else:
-            self.__index_cache[dbname][collection][index] = expire
+            else:
+                self.__index_cache[dbname][collection][index] = expire
 
     def _purge_index(self, database_name,
                      collection_name=None, index_name=None):
@@ -450,39 +486,39 @@ class MongoClient(common.BaseObject):
 
         If `collection_name` is None purge an entire database.
         """
-        if not database_name in self.__index_cache:
-            return
+        with self.__index_cache_lock:
+            if not database_name in self.__index_cache:
+                return
 
-        if collection_name is None:
-            del self.__index_cache[database_name]
-            return
+            if collection_name is None:
+                del self.__index_cache[database_name]
+                return
 
-        if not collection_name in self.__index_cache[database_name]:
-            return
+            if not collection_name in self.__index_cache[database_name]:
+                return
 
-        if index_name is None:
-            del self.__index_cache[database_name][collection_name]
-            return
+            if index_name is None:
+                del self.__index_cache[database_name][collection_name]
+                return
 
-        if index_name in self.__index_cache[database_name][collection_name]:
-            del self.__index_cache[database_name][collection_name][index_name]
+            if index_name in self.__index_cache[database_name][collection_name]:
+                del self.__index_cache[database_name][collection_name][index_name]
 
-    def _server_property(self, attr_name, default=None):
+    def _server_property(self, attr_name):
         """An attribute of the current server's description.
 
-        Returns "default" while there is no current server, primary, or mongos.
+        If the client is not connected, this will block until a connection is
+        established or raise ServerSelectionTimeoutError if no server is
+        available.
 
         Not threadsafe if used multiple times in a single method, since
         the server may change. In such cases, store a local reference to a
         ServerDescription first, then use its properties.
         """
-        try:
-            server = self._topology.select_server(
-                writable_server_selector, server_selection_timeout=0)
+        server = self._topology.select_server(
+            writable_server_selector)
 
-            return getattr(server.description, attr_name)
-        except ConnectionFailure:
-            return default
+        return getattr(server.description, attr_name)
 
     @property
     def event_listeners(self):
@@ -500,15 +536,21 @@ class MongoClient(common.BaseObject):
         the client is load-balancing among mongoses, since there is no single
         address. Use :attr:`nodes` instead.
 
+        If the client is not connected, this will block until a connection is
+        established or raise ServerSelectionTimeoutError if no server is
+        available.
+
         .. versionadded:: 3.0
         """
-        try:
-            return self._topology.get_direct_or_primary()
-        except InvalidOperation:
-            # Only one case where Topology throws InvalidOperation.
+        topology_type = self._topology._description.topology_type
+        if topology_type == TOPOLOGY_TYPE.Sharded:
             raise InvalidOperation(
                 'Cannot use "address" property when load balancing among'
                 ' mongoses, use "nodes" instead.')
+        if topology_type not in (TOPOLOGY_TYPE.ReplicaSetWithPrimary,
+                                 TOPOLOGY_TYPE.Single):
+            return None
+        return self._server_property('address')
 
     @property
     def primary(self):
@@ -550,30 +592,53 @@ class MongoClient(common.BaseObject):
 
     @property
     def is_primary(self):
-        """If this client if connected to a server that can accept writes.
+        """If this client is connected to a server that can accept writes.
 
         True if the current server is a standalone, mongos, or the primary of
-        a replica set.
+        a replica set. If the client is not connected, this will block until a
+        connection is established or raise ServerSelectionTimeoutError if no
+        server is available.
         """
-        return self._server_property('is_writable', False)
+        return self._server_property('is_writable')
 
     @property
     def is_mongos(self):
-        """If this client is connected to mongos.
+        """If this client is connected to mongos. If the client is not
+        connected, this will block until a connection is established or raise
+        ServerSelectionTimeoutError if no server is available..
         """
         return self._server_property('server_type') == SERVER_TYPE.Mongos
 
     @property
     def max_pool_size(self):
-        """The maximum number of sockets the pool will open concurrently.
+        """The maximum allowable number of concurrent connections to each
+        connected server. Requests to a server will block if there are
+        `maxPoolSize` outstanding connections to the requested server.
+        Defaults to 100. Cannot be 0.
 
-        When the pool has reached `max_pool_size`, operations block waiting for
-        a socket to be returned to the pool. If ``waitQueueTimeoutMS`` is set,
-        a blocked operation will raise :exc:`~pymongo.errors.ConnectionFailure`
-        after a timeout. By default ``waitQueueTimeoutMS`` is not set.
+        When a server's pool has reached `max_pool_size`, operations for that
+        server block waiting for a socket to be returned to the pool. If
+        ``waitQueueTimeoutMS`` is set, a blocked operation will raise
+        :exc:`~pymongo.errors.ConnectionFailure` after a timeout.
+        By default ``waitQueueTimeoutMS`` is not set.
         """
         return self.__options.pool_options.max_pool_size
 
+    @property
+    def min_pool_size(self):
+        """The minimum required number of concurrent connections that the pool
+        will maintain to each connected server. Default is 0.
+        """
+        return self.__options.pool_options.min_pool_size
+
+    @property
+    def max_idle_time_ms(self):
+        """The maximum number of milliseconds that a connection can remain
+        idle in the pool before being removed and replaced. Defaults to
+        `None` (no limit).
+        """
+        return self.__options.pool_options.max_idle_time_ms
+
     @property
     def nodes(self):
         """Set of all currently connected servers.
@@ -592,28 +657,34 @@ class MongoClient(common.BaseObject):
     def max_bson_size(self):
         """The largest BSON object the connected server accepts in bytes.
 
-        Defaults to 16MB if not connected to a server.
+        If the client is not connected, this will block until a connection is
+        established or raise ServerSelectionTimeoutError if no server is
+        available.
         """
-        return self._server_property('max_bson_size', common.MAX_BSON_SIZE)
+        return self._server_property('max_bson_size')
 
     @property
     def max_message_size(self):
         """The largest message the connected server accepts in bytes.
 
-        Defaults to 32MB if not connected to a server.
+        If the client is not connected, this will block until a connection is
+        established or raise ServerSelectionTimeoutError if no server is
+        available.
         """
-        return self._server_property(
-            'max_message_size', common.MAX_MESSAGE_SIZE)
+        return self._server_property('max_message_size')
 
     @property
     def max_write_batch_size(self):
         """The maxWriteBatchSize reported by the server.
 
+        If the client is not connected, this will block until a connection is
+        established or raise ServerSelectionTimeoutError if no server is
+        available.
+
         Returns a default value when connected to server versions prior to
         MongoDB 2.6.
         """
-        return self._server_property(
-            'max_write_batch_size', common.MAX_WRITE_BATCH_SIZE)
+        return self._server_property('max_write_batch_size')
 
     @property
     def local_threshold_ms(self):
@@ -649,7 +720,7 @@ class MongoClient(common.BaseObject):
         self._topology.close()
 
     def set_cursor_manager(self, manager_class):
-        """Set this client's cursor manager.
+        """DEPRECATED - Set this client's cursor manager.
 
         Raises :class:`TypeError` if `manager_class` is not a subclass of
         :class:`~pymongo.cursor_manager.CursorManager`. A cursor manager
@@ -660,9 +731,16 @@ class MongoClient(common.BaseObject):
         :Parameters:
           - `manager_class`: cursor manager to use
 
+        .. versionchanged:: 3.3
+           Deprecated, for real this time.
+
         .. versionchanged:: 3.0
            Undeprecated.
         """
+        warnings.warn(
+            "set_cursor_manager is Deprecated",
+            DeprecationWarning,
+            stacklevel=2)
         manager = manager_class(self)
         if not isinstance(manager, CursorManager):
             raise TypeError("manager_class must be a subclass of "
@@ -807,8 +885,8 @@ class MongoClient(common.BaseObject):
                 else:
                     return 'document_class=%s.%s' % (value.__module__,
                                                      value.__name__)
-            if "ms" in option:
-                return "%s='%s'" % (option, int(value * 1000))
+            if option in common.TIMEOUT_VALIDATORS and value is not None:
+                return "%s=%s" % (option, int(value * 1000))
 
             return '%s=%r' % (option, value)
 
@@ -857,12 +935,17 @@ class MongoClient(common.BaseObject):
         return database.Database(self, name)
 
     def close_cursor(self, cursor_id, address=None):
-        """Close a single database cursor.
+        """Send a kill cursors message soon with the given id.
 
         Raises :class:`TypeError` if `cursor_id` is not an instance of
         ``(int, long)``. What closing the cursor actually means
         depends on this client's cursor manager.
 
+        This method may be called from a :class:`~pymongo.cursor.Cursor`
+        destructor during garbage collection, so it isn't safe to take a
+        lock or do network I/O. Instead, we schedule the cursor to be closed
+        soon on a background thread.
+
         :Parameters:
           - `cursor_id`: id of cursor to close
           - `address` (optional): (host, port) pair of the cursor's server.
@@ -875,30 +958,36 @@ class MongoClient(common.BaseObject):
         if not isinstance(cursor_id, integer_types):
             raise TypeError("cursor_id must be an instance of (int, long)")
 
-        self.__cursor_manager.close(cursor_id, address)
+        if self.__cursor_manager is not None:
+            self.__cursor_manager.close(cursor_id, address)
+        else:
+            self.__kill_cursors_queue.append((address, [cursor_id]))
 
     def kill_cursors(self, cursor_ids, address=None):
-        """Send a kill cursors message soon with the given ids.
+        """DEPRECATED - Send a kill cursors message soon with the given ids.
 
         Raises :class:`TypeError` if `cursor_ids` is not an instance of
         ``list``.
 
-        This method may be called from a :class:`~pymongo.cursor.Cursor`
-        destructor during garbage collection, so it isn't safe to take a
-        lock or do network I/O. Instead, we schedule the cursor to be closed
-        soon on a background thread.
-
         :Parameters:
           - `cursor_ids`: list of cursor ids to kill
           - `address` (optional): (host, port) pair of the cursor's server.
             If it is not provided, the client attempts to close the cursor on
             the primary or standalone, or a mongos server.
 
+        .. versionchanged:: 3.3
+           Deprecated.
+
         .. versionchanged:: 3.0
            Now accepts an `address` argument. Schedules the cursors to be
            closed on a background thread instead of sending the message
            immediately.
         """
+        warnings.warn(
+            "kill_cursors is deprecated.",
+            DeprecationWarning,
+            stacklevel=2)
+
         if not isinstance(cursor_ids, list):
             raise TypeError("cursor_ids must be a list")
 
@@ -906,8 +995,9 @@ class MongoClient(common.BaseObject):
         self.__kill_cursors_queue.append((address, cursor_ids))
 
     # This method is run periodically by a background thread.
-    def _process_kill_cursors_queue(self):
-        """Process any pending kill cursors requests."""
+    def _process_periodic_tasks(self):
+        """Process any pending kill cursors requests and
+        maintain connection pool parameters."""
         address_to_cursor_ids = defaultdict(list)
 
         # Other threads or the GC may append to the queue concurrently.
@@ -936,39 +1026,55 @@ class MongoClient(common.BaseObject):
                         server = topology.select_server(
                             writable_server_selector)
 
-                    if publish:
-                        start = datetime.datetime.now()
-                    data = message.kill_cursors(cursor_ids)
-                    if publish:
-                        duration = datetime.datetime.now() - start
-                        try:
-                            dbname, collname = address.namespace.split(".", 1)
-                        except AttributeError:
-                            dbname = collname = 'OP_KILL_CURSORS'
-                        command = SON([('killCursors', collname),
-                                       ('cursors', cursor_ids)])
-                        listeners.publish_command_start(
-                            command, dbname, data[0], address)
-                        start = datetime.datetime.now()
                     try:
-                        server.send_message(data, self.__all_credentials)
-                    except Exception as exc:
-                        if publish:
-                            dur = (datetime.datetime.now() - start) + duration
-                            listeners.publish_command_failure(
-                                dur, message._convert_exception(exc),
-                                'killCursors', data[0], address)
-                        raise
-                    if publish:
-                        duration = (datetime.datetime.now() - start) + duration
-                        # OP_KILL_CURSORS returns no reply, fake one.
-                        reply = {'cursorsUnknown': cursor_ids, 'ok': 1}
-                        listeners.publish_command_success(
-                            duration, reply, 'killCursors', data[0], address)
+                        namespace = address.namespace
+                        db, coll = namespace.split('.', 1)
+                    except AttributeError:
+                        namespace = None
+                        db = coll = "OP_KILL_CURSORS"
 
-                except ConnectionFailure as exc:
-                    warnings.warn("couldn't close cursor on %s: %s"
-                                  % (address, exc))
+                    spec = SON([('killCursors', coll),
+                                ('cursors', cursor_ids)])
+                    with server.get_socket(self.__all_credentials) as sock_info:
+                        if (sock_info.max_wire_version >= 4 and
+                                namespace is not None):
+                            sock_info.command(db, spec)
+                        else:
+                            if publish:
+                                start = datetime.datetime.now()
+                            request_id, msg = message.kill_cursors(cursor_ids)
+                            if publish:
+                                duration = datetime.datetime.now() - start
+                                listeners.publish_command_start(
+                                    spec, db, request_id, address)
+                                start = datetime.datetime.now()
+
+                            try:
+                                sock_info.send_message(msg, 0)
+                            except Exception as exc:
+                                if publish:
+                                    dur = ((datetime.datetime.now() - start)
+                                           + duration)
+                                    listeners.publish_command_failure(
+                                        dur, message._convert_exception(exc),
+                                        'killCursors', request_id, address)
+                                raise
+
+                            if publish:
+                                duration = ((datetime.datetime.now() - start)
+                                            + duration)
+                                # OP_KILL_CURSORS returns no reply, fake one.
+                                reply = {'cursorsUnknown': cursor_ids, 'ok': 1}
+                                listeners.publish_command_success(
+                                    duration, reply, 'killCursors', request_id,
+                                    address)
+
+                except Exception:
+                    helpers._handle_exception()
+        try:
+            self._topology.update_pool()
+        except Exception:
+            helpers._handle_exception()
 
     def server_info(self):
         """Get information about the MongoDB server we're connected to."""
@@ -978,9 +1084,8 @@ class MongoClient(common.BaseObject):
     def database_names(self):
         """Get a list of the names of all databases on the connected server."""
         return [db["name"] for db in
-                self.admin.command(
-                    "listDatabases",
-                    read_preference=ReadPreference.PRIMARY)["databases"]]
+                self._database_default_options('admin').command(
+                    "listDatabases")["databases"]]
 
     def drop_database(self, name_or_database):
         """Drop a database.
@@ -1022,8 +1127,8 @@ class MongoClient(common.BaseObject):
 
         return self[self.__default_database_name]
 
-    def get_database(self, name, codec_options=None,
-                     read_preference=None, write_concern=None):
+    def get_database(self, name, codec_options=None, read_preference=None,
+                     write_concern=None, read_concern=None):
         """Get a :class:`~pymongo.database.Database` with the given name and
         options.
 
@@ -1056,9 +1161,21 @@ class MongoClient(common.BaseObject):
             :class:`~pymongo.write_concern.WriteConcern`. If ``None`` (the
             default) the :attr:`write_concern` of this :class:`MongoClient` is
             used.
+          - `read_concern` (optional): An instance of
+            :class:`~pymongo.read_concern.ReadConcern`. If ``None`` (the
+            default) the :attr:`read_concern` of this :class:`MongoClient` is
+            used.
         """
         return database.Database(
-            self, name, codec_options, read_preference, write_concern)
+            self, name, codec_options, read_preference,
+            write_concern, read_concern)
+
+    def _database_default_options(self, name):
+        """Get a Database instance with the default settings."""
+        return self.get_database(
+            name, codec_options=DEFAULT_CODEC_OPTIONS,
+            read_preference=ReadPreference.PRIMARY,
+            write_concern=WriteConcern())
 
     @property
     def is_locked(self):
@@ -1066,7 +1183,7 @@ class MongoClient(common.BaseObject):
         are blocked, although read operations may still be allowed.
         Use :meth:`unlock` to unlock.
         """
-        ops = self.admin.current_op()
+        ops = self._database_default_options('admin').current_op()
         return bool(ops.get('fsyncLock', 0))
 
     def fsync(self, **kwargs):
@@ -1091,9 +1208,19 @@ class MongoClient(common.BaseObject):
     def unlock(self):
         """Unlock a previously locked server.
         """
-        coll = self.admin.get_collection(
-            "$cmd.sys.unlock", read_preference=ReadPreference.PRIMARY)
-        coll.find_one()
+        cmd = {"fsyncUnlock": 1}
+        with self._socket_for_writes() as sock_info:
+            if sock_info.max_wire_version >= 4:
+                try:
+                    sock_info.command("admin", cmd)
+                except OperationFailure as exc:
+                    # Ignore "DB not locked" to replicate old behavior
+                    if exc.code != 125:
+                        raise
+            else:
+                helpers._first_batch(sock_info, "admin", "$cmd.sys.unlock",
+                    {}, -1, True, self.codec_options,
+                    ReadPreference.PRIMARY, cmd, self._event_listeners)
 
     def __enter__(self):
         return self
